@@ -13,7 +13,9 @@
 import collections
 import contextlib
 import copy
+import fixtures
 import pickle
+import sys
 import warnings
 
 import mock
@@ -24,6 +26,7 @@ from sqlalchemy import Column
 from sqlalchemy import Integer
 from sqlalchemy import MetaData
 from sqlalchemy.orm import mapper
+from sqlalchemy.orm import Session
 from sqlalchemy import select
 from sqlalchemy import String
 from sqlalchemy import Table
@@ -33,7 +36,7 @@ from oslo_db import options
 from oslo_db.sqlalchemy import enginefacade
 from oslo_db.sqlalchemy import engines as oslo_engines
 from oslo_db.sqlalchemy import orm
-from oslo_db.sqlalchemy import test_base
+from oslo_db.tests.sqlalchemy import base as test_base
 
 
 enginefacade.transaction_context_provider(oslo_context.RequestContext)
@@ -55,6 +58,7 @@ class SingletonConnection(SingletonOnName):
     def __init__(self, **kw):
         super(SingletonConnection, self).__init__(
             "connection", **kw)
+        self.info = {}
 
 
 class SingletonEngine(SingletonOnName):
@@ -113,14 +117,16 @@ class MockFacadeTest(oslo_test_base.BaseTestCase):
         writer_conn = SingletonConnection()
         writer_engine = SingletonEngine(writer_conn)
         writer_session = mock.Mock(
-            connection=mock.Mock(return_value=writer_conn))
+            connection=mock.Mock(return_value=writer_conn),
+            info={})
         writer_maker = mock.Mock(return_value=writer_session)
 
         if self.slave_uri:
             async_reader_conn = SingletonConnection()
             async_reader_engine = SingletonEngine(async_reader_conn)
             async_reader_session = mock.Mock(
-                connection=mock.Mock(return_value=async_reader_conn))
+                connection=mock.Mock(return_value=async_reader_conn),
+                info={})
             async_reader_maker = mock.Mock(return_value=async_reader_session)
 
         else:
@@ -130,15 +136,15 @@ class MockFacadeTest(oslo_test_base.BaseTestCase):
             async_reader_maker = writer_maker
 
         if self.synchronous_reader:
-            reader_conn = async_reader_conn
-            reader_engine = async_reader_engine
-            reader_session = async_reader_session
-            reader_maker = async_reader_maker
-        else:
             reader_conn = writer_conn
             reader_engine = writer_engine
             reader_session = writer_session
             reader_maker = writer_maker
+        else:
+            reader_conn = async_reader_conn
+            reader_engine = async_reader_engine
+            reader_session = async_reader_session
+            reader_maker = async_reader_maker
 
         self.connections = AssertDataSource(
             writer_conn, reader_conn, async_reader_conn
@@ -231,9 +237,9 @@ class MockFacadeTest(oslo_test_base.BaseTestCase):
             async_reader_engine = writer_engine
 
         if self.synchronous_reader:
-            reader_engine = async_reader_engine
-        else:
             reader_engine = writer_engine
+        else:
+            reader_engine = async_reader_engine
 
         engines = AssertDataSource(
             writer_engine, reader_engine, async_reader_engine)
@@ -328,9 +334,9 @@ class MockFacadeTest(oslo_test_base.BaseTestCase):
             async_reader_maker = writer_maker
 
         if self.synchronous_reader:
-            reader_maker = async_reader_maker
-        else:
             reader_maker = writer_maker
+        else:
+            reader_maker = async_reader_maker
 
         makers = AssertDataSource(
             writer_maker,
@@ -389,6 +395,12 @@ class MockFacadeTest(oslo_test_base.BaseTestCase):
             makers, enginefacade._WRITER,
             connection, assert_calls)
 
+    def _emit_sub_writer_session(self, session):
+        return self._emit_sub_session(enginefacade._WRITER, session)
+
+    def _emit_sub_reader_session(self, session):
+        return self._emit_sub_session(enginefacade._READER, session)
+
     @contextlib.contextmanager
     def _assert_session(
             self, makers, writer, connection=None, assert_calls=True):
@@ -418,6 +430,13 @@ class MockFacadeTest(oslo_test_base.BaseTestCase):
                 session.mock_calls,
                 self.sessions.element_for_writer(writer).mock_calls)
 
+    @contextlib.contextmanager
+    def _emit_sub_session(self, writer, session):
+        yield session
+        if enginefacade._context_manager.\
+                _factory._transaction_ctx_cfg['flush_on_subtransaction']:
+            session.flush()
+
     def test_dispose_pool(self):
         facade = enginefacade.transaction_context()
 
@@ -589,7 +608,7 @@ class MockFacadeTest(oslo_test_base.BaseTestCase):
 
     def test_async_on_writer_raises(self):
         exc = self.assertRaises(
-            TypeError, getattr, enginefacade.writer, "async"
+            TypeError, getattr, enginefacade.writer, "async_"
         )
         self.assertEqual(
             "Setting async on a WRITER makes no sense",
@@ -608,7 +627,7 @@ class MockFacadeTest(oslo_test_base.BaseTestCase):
     def test_reader_nested_in_async_reader_raises(self):
         context = oslo_context.RequestContext()
 
-        @enginefacade.reader.async
+        @enginefacade.reader.async_
         def go1(context):
             context.session.execute("test1")
             go2(context)
@@ -629,7 +648,7 @@ class MockFacadeTest(oslo_test_base.BaseTestCase):
     def test_reader_allow_async_nested_in_async_reader(self):
         context = oslo_context.RequestContext()
 
-        @enginefacade.reader.async
+        @enginefacade.reader.async_
         def go1(context):
             context.session.execute("test1")
             go2(context)
@@ -683,7 +702,7 @@ class MockFacadeTest(oslo_test_base.BaseTestCase):
     def test_writer_nested_in_async_reader_raises(self):
         context = oslo_context.RequestContext()
 
-        @enginefacade.reader.async
+        @enginefacade.reader.async_
         def go1(context):
             context.session.execute("test1")
             go2(context)
@@ -723,10 +742,30 @@ class MockFacadeTest(oslo_test_base.BaseTestCase):
                 with self._assert_writer_session(makers) as session:
                     session.execute("test2")
 
+    def test_deprecated_async_reader_name(self):
+        if sys.version_info >= (3, 7):
+            self.skipTest("Test only runs on Python < 3.7")
+
+        context = oslo_context.RequestContext()
+
+        old = getattr(enginefacade.reader, "async")
+
+        @old
+        def go1(context):
+            context.session.execute("test1")
+
+        go1(context)
+
+        with self._assert_engines() as engines:
+            with self._assert_makers(engines) as makers:
+                with self._assert_async_reader_session(
+                        makers, assert_calls=False) as session:
+                    session.execute("test1")
+
     def test_async_reader_then_reader_ok(self):
         context = oslo_context.RequestContext()
 
-        @enginefacade.reader.async
+        @enginefacade.reader.async_
         def go1(context):
             context.session.execute("test1")
 
@@ -757,6 +796,20 @@ class MockFacadeTest(oslo_test_base.BaseTestCase):
                 with self._assert_reader_session(makers) as session:
                     session.execute("test1")
 
+    def test_using_context_present_in_session_info(self):
+        context = oslo_context.RequestContext()
+
+        with enginefacade.reader.using(context) as session:
+            self.assertEqual(context, session.info['using_context'])
+        self.assertIsNone(session.info['using_context'])
+
+    def test_using_context_present_in_connection_info(self):
+        context = oslo_context.RequestContext()
+
+        with enginefacade.writer.connection.using(context) as connection:
+            self.assertEqual(context, connection.info['using_context'])
+        self.assertIsNone(connection.info['using_context'])
+
     def test_using_reader_rollback_reader_session(self):
         enginefacade.configure(rollback_reader_sessions=True)
 
@@ -771,6 +824,23 @@ class MockFacadeTest(oslo_test_base.BaseTestCase):
                 with self._assert_reader_session(makers) as session:
                     session.execute("test1")
 
+    def test_using_flush_on_nested(self):
+        enginefacade.configure(flush_on_nested=True)
+
+        context = oslo_context.RequestContext()
+
+        with enginefacade.writer.using(context) as session:
+            with enginefacade.writer.using(context) as session:
+                self._assert_ctx_session(context, session)
+                session.execute("test1")
+
+        with self._assert_engines() as engines:
+            with self._assert_makers(engines) as makers:
+                with self._assert_writer_session(makers) as session:
+                    with self._emit_sub_writer_session(
+                            session) as session:
+                        session.execute("test1")
+
     def test_using_writer(self):
         context = oslo_context.RequestContext()
 
@@ -877,7 +947,7 @@ class MockFacadeTest(oslo_test_base.BaseTestCase):
                 getattr, context, 'session'
             )
 
-            self.assertRegexpMatches(
+            self.assertRegex(
                 exc.args[0],
                 "The 'session' context attribute was requested but it has "
                 "not been established for this context."
@@ -892,7 +962,7 @@ class MockFacadeTest(oslo_test_base.BaseTestCase):
                 getattr, context, 'connection'
             )
 
-            self.assertRegexpMatches(
+            self.assertRegex(
                 exc.args[0],
                 "The 'connection' context attribute was requested but it has "
                 "not been established for this context."
@@ -905,7 +975,7 @@ class MockFacadeTest(oslo_test_base.BaseTestCase):
             getattr, context, 'session'
         )
 
-        self.assertRegexpMatches(
+        self.assertRegex(
             exc.args[0],
             "No TransactionContext is established for "
             "this .*RequestContext.* object within the current "
@@ -923,7 +993,7 @@ class MockFacadeTest(oslo_test_base.BaseTestCase):
             getattr, context, 'connection'
         )
 
-        self.assertRegexpMatches(
+        self.assertRegex(
             exc.args[0],
             "No TransactionContext is established for "
             "this .*RequestContext.* object within the current "
@@ -941,7 +1011,7 @@ class MockFacadeTest(oslo_test_base.BaseTestCase):
             getattr, context, 'transaction'
         )
 
-        self.assertRegexpMatches(
+        self.assertRegex(
             exc.args[0],
             "No TransactionContext is established for "
             "this .*RequestContext.* object within the current "
@@ -959,7 +1029,7 @@ class MockFacadeTest(oslo_test_base.BaseTestCase):
             getattr, context, 'transaction_ctx'
         )
 
-        self.assertRegexpMatches(
+        self.assertRegex(
             exc.args[0],
             "No TransactionContext is established for "
             "this .*RequestContext.* object within the current "
@@ -1160,7 +1230,10 @@ class PatchFactoryTest(oslo_test_base.BaseTestCase):
 
     def test_patch_engine(self):
         normal_mgr = enginefacade.transaction_context()
-        normal_mgr.configure(connection="sqlite:///foo.db")
+        normal_mgr.configure(
+            connection="sqlite:///foo.db",
+            rollback_reader_sessions=True
+        )
 
         @normal_mgr.writer
         def go1(context):
@@ -1179,6 +1252,17 @@ class PatchFactoryTest(oslo_test_base.BaseTestCase):
                 s1.bind.url,
                 "sqlite:///bar.db")
 
+            self.assertTrue(
+                enginefacade._transaction_ctx_for_context(
+                    context).rollback_reader_sessions
+            )
+
+            # ensure this defaults to True
+            self.assertTrue(
+                enginefacade._transaction_ctx_for_context(
+                    context).factory.synchronous_reader
+            )
+
         def create_engine(sql_connection, **kw):
             return mock.Mock(url=sql_connection)
 
@@ -1195,6 +1279,40 @@ class PatchFactoryTest(oslo_test_base.BaseTestCase):
             reset()
             go1(context)
 
+    def test_patch_not_started(self):
+        normal_mgr = enginefacade.transaction_context()
+        normal_mgr.configure(
+            connection="sqlite:///foo.db",
+            rollback_reader_sessions=True
+        )
+
+        @normal_mgr.writer
+        def go1(context):
+            s1 = context.session
+
+            self.assertEqual(
+                s1.bind.url,
+                "sqlite:///bar.db")
+
+            self.assertTrue(
+                enginefacade._transaction_ctx_for_context(
+                    context).rollback_reader_sessions
+            )
+
+        def create_engine(sql_connection, **kw):
+            return mock.Mock(url=sql_connection)
+
+        with mock.patch(
+                "oslo_db.sqlalchemy.engines.create_engine", create_engine):
+            mock_engine = create_engine("sqlite:///bar.db")
+
+            context = oslo_context.RequestContext()
+            reset = normal_mgr.patch_engine(mock_engine)
+            go1(context)
+            self.assertIs(
+                normal_mgr._factory._writer_engine, mock_engine)
+            reset()
+
     def test_new_manager_from_config(self):
         normal_mgr = enginefacade.transaction_context()
         normal_mgr.configure(
@@ -1238,7 +1356,7 @@ class AsyncReaderWSlaveMockFacadeTest(MockFacadeTest):
     slave_uri = 'some_slave_connection'
 
 
-class LegacyIntegrationtest(test_base.DbTestCase):
+class LegacyIntegrationtest(test_base._DbTestCase):
 
     def test_legacy_integration(self):
         legacy_facade = enginefacade.get_legacy_facade()
@@ -1296,7 +1414,7 @@ class LegacyIntegrationtest(test_base.DbTestCase):
         )
 
 
-class ThreadingTest(test_base.DbTestCase):
+class ThreadingTest(test_base._DbTestCase):
     """Test copy/pickle on new threads using real connections and sessions."""
 
     def _assert_ctx_connection(self, context, connection):
@@ -1489,7 +1607,7 @@ class ThreadingTest(test_base.DbTestCase):
                 assert session is not session2
 
 
-class LiveFacadeTest(test_base.DbTestCase):
+class LiveFacadeTest(test_base._DbTestCase):
     """test using live SQL with test-provisioned databases.
 
     Several of these tests require that multiple transactions run
@@ -1507,6 +1625,7 @@ class LiveFacadeTest(test_base.DbTestCase):
             'user', metadata,
             Column('id', Integer, primary_key=True),
             Column('name', String(30)),
+            Column('favorite_color', String(10), default='yellow'),
             mysql_engine='InnoDB'
         )
         self.user_table = user_table
@@ -1558,6 +1677,57 @@ class LiveFacadeTest(test_base.DbTestCase):
             session.query(self.User.name).scalar()
         )
 
+    @mock.patch.object(Session, 'commit')
+    @mock.patch.object(Session, 'rollback')
+    def test_save_and_reraise_when_rollback_exception(self,
+                                                      rollback_patch,
+                                                      commit_patch):
+        context = oslo_context.RequestContext()
+        log = self.useFixture(fixtures.FakeLogger())
+
+        class RollbackException(Exception):
+            pass
+
+        class CommitException(Exception):
+            pass
+
+        commit_patch.side_effect = CommitException()
+        rollback_patch.side_effect = RollbackException()
+
+        @enginefacade.writer
+        def go_session(context):
+            context.session.add(self.User(name="u1"))
+
+        self.assertRaises(RollbackException, go_session, context)
+        self.assertIn('CommitException', log.output)
+
+    def test_flush_on_subtransaction(self):
+        facade = enginefacade.transaction_context()
+        facade.configure(
+            connection=self.engine.url,
+            flush_on_subtransaction=True)
+        facade.patch_engine(self.engine)
+        context = oslo_context.RequestContext()
+
+        with facade.writer.using(context):
+            with facade.writer.using(context):
+                u = self.User(name="u1")
+                context.session.add(u)
+            self.assertEqual('yellow', u.favorite_color)
+
+    def test_flush_on_subtransaction_default_off(self):
+        context = oslo_context.RequestContext()
+        facade = enginefacade.transaction_context()
+        facade.configure(connection=self.engine.url)
+        facade.patch_engine(self.engine)
+
+        with facade.writer.using(context):
+            with facade.writer.using(context):
+                u = self.User(name="u1")
+                context.session.add(u)
+            self.assertIsNone(u.favorite_color)
+        self.assertEqual('yellow', u.favorite_color)
+
     def test_context_deepcopy_on_session(self):
         context = oslo_context.RequestContext()
         with enginefacade.writer.using(context) as session:
@@ -1631,7 +1801,7 @@ class LiveFacadeTest(test_base.DbTestCase):
 
         session = self.sessionmaker(autocommit=True)
 
-        # inner transction + second part of "outer" transaction were committed
+        # inner transaction + second part of "outer" transaction were committed
         self.assertEqual(
             [("u2",), ("u3",), ("u4", )],
             session.query(
@@ -1671,7 +1841,7 @@ class LiveFacadeTest(test_base.DbTestCase):
 
         session = self.sessionmaker(autocommit=True)
 
-        # inner transction + second part of "outer" transaction were committed
+        # inner transaction + second part of "outer" transaction were committed
         self.assertEqual(
             [("u1",), ("u3",), ("u4", )],
             session.query(
@@ -1699,7 +1869,7 @@ class LiveFacadeTest(test_base.DbTestCase):
 
         session = self.sessionmaker(autocommit=True)
 
-        # inner transction + second part of "outer" transaction were committed
+        # inner transaction + second part of "outer" transaction were committed
         self.assertEqual(
             [("u1",), ("u3",), ("u4", )],
             session.query(
@@ -1747,7 +1917,7 @@ class LiveFacadeTest(test_base.DbTestCase):
 
         session = self.sessionmaker(autocommit=True)
 
-        # inner transction + second part of "outer" transaction were committed
+        # inner transaction + second part of "outer" transaction were committed
         self.assertEqual(
             [("u2",), ("u3",), ("u4", )],
             session.query(
@@ -1978,12 +2148,12 @@ class LiveFacadeTest(test_base.DbTestCase):
 
 
 class MySQLLiveFacadeTest(
-        test_base.MySQLOpportunisticTestCase, LiveFacadeTest):
+        test_base._MySQLOpportunisticTestCase, LiveFacadeTest):
     pass
 
 
 class PGLiveFacadeTest(
-        test_base.PostgreSQLOpportunisticTestCase, LiveFacadeTest):
+        test_base._PostgreSQLOpportunisticTestCase, LiveFacadeTest):
     pass
 
 
@@ -2007,7 +2177,8 @@ class ConfigOptionsTest(oslo_test_base.BaseTestCase):
         with warnings.catch_warnings(record=True) as w:
             warnings.simplefilter("always")
 
-            factory.configure(fake1='x', idle_timeout=200, wrong2='y')
+            factory.configure(
+                fake1='x', connection_recycle_time=200, wrong2='y')
 
         self.assertEqual(1, len(w))
         self.assertTrue(
